# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import platform
import re
import subprocess
import tempfile
from pathlib import Path

# Related third-party imports
import lxml.etree
import lxml.html
from mistletoe import Document
from mistletoe.block_token import ThematicBreak
from mistletoe.markdown_renderer import MarkdownRenderer
from mistletoe.span_token import Image

from wurzel.exceptions import InvalidPlatform, MarkdownConvertFailed

# pylint: disable=c-extension-no-member


def __get_html2md() -> Path:
    default_path = {
        "Linux_x86_64": Path(__file__).parent / "html2md",
        "Darwin_arm64": Path(__file__).parent / "html2md_darwin_arm",
        "Darwin_x86_64": Path(__file__).parent / "html2md_darwin_amd64",
    }
    fallback = default_path.get(
        f"{platform.uname().system}_{platform.uname().machine}", None
    )
    if fallback is None:
        raise InvalidPlatform(
            f"Could not create path to binary from {platform.uname()} we only support {default_path.keys()}"
        )
    return Path(fallback)


__HTML2MD: Path = __get_html2md()


"""
Wrapper module around html2md binary
"""


# mypy: ignore-errors


def to_markdown(html: str, binary_path: Path = __HTML2MD) -> str:
    """
    Convert HTML XML string to Markdown using an external binary or a Python library.
    In acknowledge to https://github.com/suntong/html2md
    Parameters
    ----------
    html : str
        The input HTML/XML string to be converted to Markdown.
    binary_path : Path, optional
        The path to the html2md binary (default is './html2md').

    Returns
    -------
    str
        The resulting Markdown string.

    Notes
    -----
    This function first checks if the html2md binary is available. If not, it raises an error.
    The binary path can be specified in the environment variable 'HTML2MD_BINARY_PATH'.

    The html2md binary can be found at: https://github.com/suntong/html2md

    Examples
    --------
    >>> from pathlib import Path
    >>> html = '<h1>Title</h1><p>Hello, world!</p>'
    >>> markdown = to_markdown(html)
    >>> print(markdown)
    # Title
    Hello, world!
    """
    with tempfile.NamedTemporaryFile(delete=False, suffix=".html", mode="w+") as file:
        cleaned_html = clean_html(html)
        file.write(cleaned_html)
        file.close()
        convert_cmd = f"{binary_path.absolute().as_posix()} -i {file.name}"
        status_code, markdown = subprocess.getstatusoutput(convert_cmd)
        Path(file.name).unlink()
    if status_code != 0:
        raise MarkdownConvertFailed(
            f"{binary_path} returned {status_code} ({markdown} based on {html})"
        )
    if not markdown.replace(" ", "").replace("\n", ""):
        raise MarkdownConvertFailed(f"Failed to convert {html} to md {markdown}")
    return markdown


def remove_images(markdown: str) -> str:
    """
    Recursively remove image and thematic break tokens from a document object.

    Parameters
    ----------
    doc : Document
        The document object from which image and thematic break tokens are to be removed.

    Returns
    -------
    Document
        The modified document object without image and thematic break tokens.
    """

    def _to_markdown(doc: Document) -> str:
        with MarkdownRenderer() as renderer:
            rendered = renderer.render(doc)
        # Adjust for excessive newlines
        return re.sub(r"\n\n+", "\n\n", rendered)

    def _remove_image_from_document(doc: Document) -> Document:
        if not hasattr(doc, "children") or not doc.children:
            return doc
        doc.children = [
            _remove_image_from_document(x)
            for x in doc.children
            if not isinstance(x, (Image, ThematicBreak))
        ]
        return doc

    doc = Document(markdown)
    cleaned = _remove_image_from_document(doc)
    return _to_markdown(cleaned)


def clean_tree(div: lxml.etree.ElementBase) -> lxml.etree.ElementBase:
    """cleans the lxml.html tree from html unneded html obstacales"""
    # Remove all link or script tags
    for tag in ["script", "link", "style", "svg", "footer"]:
        for bad in div.xpath("//" + tag):
            bad.getparent().remove(bad)
    # replace img with it's alt text
    for img in div.xpath("//img"):
        alt_text = img.get("alt", "")
        # Create a text node with the alt text
        text_node = lxml.html.Element("span")
        text_node.text = alt_text
        # Replace the img tag with the text node
        img.addnext(text_node)
        img.getparent().remove(img)

    for bad in div.xpath("//div[@id='js-footer']"):
        bad.getparent().remove(bad)

    return div


def clean_html(html: str) -> str:
    """Clean HTML string"""
    tree = lxml.html.fromstring(html)
    cleaned_tree = clean_tree(tree)
    cleaned_html = lxml.html.tostring(cleaned_tree).decode()
    return cleaned_html


def normalize_urls_in_tree(
    tree: lxml.html.HtmlElement, base_url: str = "https://www.magenta.at"
):
    """
    Normalizes all relative URLs within an lxml HTML tree by converting them to absolute URLs.

    This function searches through the parsed HTML tree (`tree`) for elements that contain
    `href` or `src` attributes (commonly found in `<a>`, `<img>`, `<link>`, etc.), and if the value
    of these attributes is a relative URL (starting with `/`), it replaces the relative path with an
    absolute URL based on the provided `base_url`.

    Args:
        tree (lxml.html.HtmlElement): The root element of the parsed HTML tree.
        base_url (str, optional): The base URL to be used for converting relative URLs to absolute URLs.
                                  Defaults to "https://www.magenta.at".

    Example:
        If an element has a `href` or `src` like "/faq", it will be replaced with "https://www.magenta.at/faq".

    Returns:
        None: The function modifies the HTML tree in place.
    """

    # List of attributes to check for relative URLs
    attributes = ["href", "src"]

    # Loop over all elements that could have relative URLs
    for element in tree.xpath("//*[@href or @src]"):
        for attr in attributes:
            url = element.get(attr)
            if url and url.startswith("/"):
                # Replace with the absolute URL
                element.set(attr, base_url + url)
    return tree


def normalize_urls(html_content: str, base_url: str = "https://www.magenta.at"):
    """
    Converts all relative URLs in the provided HTML content to absolute URLs.

    This function parses the input HTML content, searches for elements with
    attributes that typically contain URLs (such as `href` and `src`), and
    replaces any URLs that start with a relative path (e.g., `/path`) with
    an absolute URL. The base URL used for this transformation is specified
    by the `base_url` parameter, which defaults to "https://www.magenta.at".

    Args:
        html_content (str): The HTML content as a string where relative URLs
                            need to be normalized.
        base_url (str): The base URL to prepend to relative URLs. Defaults to
                        "https://www.magenta.at".

    Returns:
        str: The HTML content with all relative URLs replaced by absolute URLs.
    """
    tree = lxml.html.fromstring(html_content)
    tree = normalize_urls_in_tree(tree, base_url)
    return html2str(tree)


def html2str(html: lxml) -> str:
    """converts lxml html to str"""
    return lxml.html.tostring(html, pretty_print=False, method="html").decode("utf-8")



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0


from typing import Generic, Optional, get_args

from wurzel.step import TypedStep
from wurzel.step.typed_step import OUTCONTRACT, SETTS


class SelfConsumingLeafStep(
    TypedStep[SETTS, OUTCONTRACT, OUTCONTRACT], Generic[SETTS, OUTCONTRACT]
):
    """Some use cases require self awareness about their last results to reduce double work."""

    def run(self, inpt: Optional[OUTCONTRACT]) -> OUTCONTRACT:
        return super().run(inpt)

    @classmethod
    def _prepare_datamodels(cls, type_annotations):
        cls.settings_class, cls.output_model_type = type_annotations
        cls.input_model_type = cls.output_model_type | None

    @classmethod
    def _prepare_instance_datamodels(cls, instance):
        instance.input_model_class = instance.output_model_class = (
            get_args(instance.output_model_type) or [instance.output_model_type]
        )[-1]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import json
from typing import Optional

from pydantic import Field, validator
from qdrant_client.models import Distance

from wurzel.step.settings import Settings


# pylint: disable=duplicate-code
class QdrantSettings(Settings):
    """All Settings related to Qdrant DB"""

    DISTANCE: Distance = Distance.DOT
    URI: str = "http://localhost:6333"
    COLLECTION: str
    COLLECTION_HISTORY_LEN: int = 10
    SEARCH_PARAMS: dict = {"metric_type": "IP", "params": {}}
    INDEX_PARAMS: dict = {
        "index_type": "FLAT",
        "field_name": "vector",
        "distance": "Dot",
        "params": {},
    }
    APIKEY: Optional[str] = ""
    REPLICATION_FACTOR: int = Field(default=3, gt=0)
    BATCH_SIZE: int = Field(default=1024, gt=0)

    @validator("SEARCH_PARAMS", "INDEX_PARAMS", pre=True)
    @classmethod
    def parse_json(cls, v):
        """validation for json"""
        if isinstance(v, str):
            return json.loads(v)
        return v



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Any, Generic, Type, TypeVar, get_args

import pandera.typing as patyp
import pydantic
from pydantic_core import PydanticCustomError, core_schema

B = TypeVar(
    "B",
    pydantic.BaseModel,
    None,
    patyp.DataFrame,
    list[pydantic.BaseModel],
    list[patyp.DataFrame],
    list[None],
)


class PathToFolderWithBaseModels(type(Path()), Generic[B]):  # type: ignore[misc]
    """Used to Store both path information
    as well as pydantic.BaseModel info.

    Inherits from Pathlib.Path:
    - Supports all path operations like / etc.


    #### Example:
    ```python
        class Person(pydantic.BaseModel):
            # Could be anything that inherits from BaseModel
            name: str
            age: int
        # Define new class to supply GenericType
        class JFP(PathToBaseModel[Person]): # JsonFilePath
            pass

        path = JFP("./max.json")
        # Load from path
        p_max = path.load_model()
        # Do anything
        p_max.age = p_max.age += 1
        # Store at path
        # Actual path can be modified like pathlib.Path
        path.save_model(p_max)
    ```
    """

    @classmethod
    def _type(cls) -> Type["PathToFolderWithBaseModels"]:
        """Get own type (used for pydantic)

        Raises:
            RuntimeError: if own type can't be found

        Returns:
            Type: PathToBaseModel
        """
        typ = getattr(cls, "__orig_bases__", [None])[0]
        if typ is None:
            raise RuntimeError(f"Could not get __orig_bases__ from class '{cls}'")
        return typ

    @classmethod
    def model_type(cls) -> Type[B]:
        """_summary_

        Raises:
            RuntimeError: if type can't be found
        Returns:
            Type[pydantic.BaseModel]: Type of Generic
        """
        try:
            return get_args(cls._type())[0]
        except Exception as err:
            raise RuntimeError("Model type could not be found") from err

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _: Any, handler: pydantic.GetCoreSchemaHandler
    ):
        return core_schema.with_info_after_validator_function(
            cls._validate_path, handler(cls._type())
        )

    @classmethod
    def _validate_path(
        cls, path: "PathToFolderWithBaseModels", _: core_schema.ValidationInfo
    ) -> Path:
        if not path.is_dir():
            raise PydanticCustomError(
                "path_not_dir", "Path is not a directory but a file"
            )
        return cls(path)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from .huggingface import HuggingFaceInferenceAPIEmbeddings, PrefixedAPIEmbeddings
from .step import EmbeddingStep

__all__ = [
    "EmbeddingStep",
    "HuggingFaceInferenceAPIEmbeddings",
    "PrefixedAPIEmbeddings",
]



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import re
from pathlib import Path
from typing import Annotated

from pydantic import Field, WrapValidator
from pydantic_core import Url

from wurzel.steps.splitter import SplitterSettings


class EmbeddingSettings(SplitterSettings):
    """Anything Embedding-related"""

    @staticmethod
    def _wrap_validator_model_mapping(input_dict: dict[str, str], handler):
        new_dict = {}
        for regex, prefix in input_dict.items():
            if isinstance(regex, str):
                new_dict[re.compile(regex)] = prefix
            else:
                new_dict.update({regex: prefix})
        return handler(new_dict)

    API: Url
    NORMALIZE: bool = False
    BATCH_SIZE: int = Field(100, gt=0)
    TOKEN_COUNT_MIN: int = Field(64, gt=0)
    TOKEN_COUNT_MAX: int = Field(256, gt=1)
    TOKEN_COUNT_BUFFER: int = Field(32, gt=0)
    STEPWORDS_PATH: Path = Path("data/german_stopwords_full.txt")
    N_JOBS: int = Field(1, gt=0)
    PREFIX_MAP: Annotated[
        dict[re.Pattern, str], WrapValidator(_wrap_validator_model_mapping)
    ] = Field(default={"e5-": "query : ", "DPR|dpr": ""})



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pandera import Field
from pandera.typing import Series

from wurzel.datacontract import PanderaDataFrameModel


class EmbeddingResult(PanderaDataFrameModel):
    """data contract of embedding output"""

    text: Series[str]
    url: Series[str] = Field(
        nullable=True, default=" ", coerce=True, description="url used for search "
    )
    vector: Series[list[float]]
    keywords: Series[str] = Field(
        nullable=True, default=" ", coerce=True, description="Keywords used for search "
    )


class EmbeddingMultiVectorResult(PanderaDataFrameModel):
    """data contract of embedding output"""

    text: Series[str]
    url: Series[str] = Field(
        nullable=True, default=" ", coerce=True, description="url used for search "
    )
    vectors: Series[list[list[float]]]
    keywords: Series[str] = Field(
        nullable=True, default=" ", coerce=True, description="Keywords used for search "
    )
    splits: Series[list[str]] = Field(
        nullable=True,
        default=None,
        coerce=True,
        description="splits of the Multivector",
    )



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Type

from wurzel.step import TypedStep


def main(
    step: Type[TypedStep],
    output_path: Path,
    input_folders: set[Path],
    executor_str_value: str,
    encapsulate_env=True,
):
    """Execute"""
    executor = executor_str_value
    with executor(dont_encapsulate=not encapsulate_env) as ex:
        ex(step, input_folders, output_path)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import inspect
from pathlib import Path
from typing import Any, Type, TypedDict

import yaml

import wurzel
import wurzel.cli
from wurzel.step import TypedStep
from wurzel.step_executor import BaseStepExecutor, PrometheusStepExecutor


class DvcDict(TypedDict):
    """Internal Representation"""

    cmd: str
    deps: list[str]
    outs: list[str]
    always_changed: bool


class Backend:
    """Abstract class to specify the Backend"""

    def generate_dict(self, step: TypedStep):
        """generate the dict"""
        raise NotImplementedError()

    def generate_yaml(self, step: TypedStep):
        """generate the dict and saves it"""
        raise NotImplementedError()


class DvcBackend(Backend):
    """'Adapter' which creates DVC yamls"""

    def __init__(
        self,
        data_dir: Path = Path("./data"),
        executer: BaseStepExecutor = PrometheusStepExecutor,
        encapsulate_env: bool = True,
    ) -> None:
        if not isinstance(data_dir, Path):
            data_dir = Path(data_dir)
        self.executor: Type[BaseStepExecutor] = executer
        self.data_dir = data_dir
        self.encapsulate_env = encapsulate_env
        super().__init__()

    def generate_dict(
        self,
        step: TypedStep,
    ) -> dict[str, DvcDict]:
        """generates the resulting dvc.yaml as dict by calling all
        its required steps as well, in recursive manner.

        Returns
        -------
        dict
            _description_
        """

        result: dict[str, Any] = {}
        outputs_of_deps: list[Path] = []
        for o_step in step.required_steps:
            dep_result = self.generate_dict(o_step)
            result |= dep_result
            outputs_of_deps += dep_result[o_step.__class__.__name__]["outs"]
        output_path = self.data_dir / step.__class__.__name__
        cmd = wurzel.cli.generate_cli_call(
            step.__class__,
            inputs=outputs_of_deps,
            output=output_path,
            executor=self.executor,
            encapsulate_env=self.encapsulate_env,
        )
        return result | {
            step.__class__.__name__: {
                "cmd": cmd,
                "deps": [inspect.getfile(step.__class__), *outputs_of_deps],
                "outs": [output_path],
                "always_changed": step.is_leaf(),
            }
        }

    def generate_yaml(
        self,
        step: TypedStep,
    ) -> str:
        """generates the dvc.yaml"""
        data = self.generate_dict(step)
        for k in data:
            for key in ["outs", "deps"]:
                data[k][key] = [str(p) for p in data[k][key]]
        return yaml.dump({"stages": data})

    @classmethod
    def save_yaml(cls, yml: str, file: Path):
        """saves given yml string int file"""
        with open(file, "w", encoding="utf-8") as f:
            yaml.dump(yml, f)



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import logging.config

from wurzel.utils.logging import JsonFormatter


class WithExtraFormatter(JsonFormatter):
    """Custom formatter with some structured logging support"""

    def format(self, record: logging.LogRecord) -> str:
        super().format(record)
        json_part = self._get_output_dict(record)
        msg = json_part.pop("message")
        if self.reduced_levels and record.levelno in self.reduced_levels:
            json_part.pop("thread", None)
            json_part.pop("threadName", None)
            json_part.pop("processName", None)
            json_part.pop("process", None)
        json_part.pop("level", None)
        json_part.pop("@timestamp", None)
        json_part.pop("file", None)
        exc_text = json_part.pop("exc_text", "")
        return " ".join(
            [
                f"'{msg}'"
                + (f" : {json.dumps(json_part)}" if json_part else "")
                + exc_text
            ]
        )



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import abc
import glob
from logging import getLogger
from pathlib import Path
from re import Pattern
from typing import Set

log = getLogger(__name__)


class Step(abc.ABC):
    """abstract class to define a WurzelStep"""

    required_steps: Set["Step"]

    def __init__(self) -> None:  # pylint: disable=dangerous-default-value
        self.required_steps = set()

    def _search_needed_files(
        self, inputs: Set[Path], file_name_pattern: Pattern
    ) -> Set[Path]:
        matched_files = set()
        for folder in inputs:
            match = set(
                glob.glob(f"{str(folder)}/**/*{str(file_name_pattern)}*")
            ) | set(glob.glob(f"{str(folder)}/*{str(file_name_pattern)}*"))
            if not match:
                log.warning(
                    f"passed path {folder} but did not find a matching file {file_name_pattern}"
                )
            matched_files |= match

        return set(map(Path, matched_files))

    def is_leaf(self) -> bool:
        """returns if the node is a leaf in the pipeline tree

        Returns
        -------
        bool
            _description_
        """
        return not self.required_steps

    def output_path(self, folder: Path) -> Path:
        """Returns a generated output path"""
        return (folder / self.__class__.__name__).with_suffix(".file")

    def add_required_step(self, step: "Step"):
        """adds a required step

        Args:
            step (Step): will be added to required_steps
        """
        self.required_steps |= {step}

    def __rshift__(self, step: "Step"):
        # Maybe Todo:
        # Check if classname is already present
        # Disallow Cyclic graphs !
        step.add_required_step(self)
        return step



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

import json

from pydantic import Field, validator

from wurzel.step.settings import Settings


# pylint: disable=duplicate-code
class MilvusSettings(Settings):
    """All Settings related to MivusDB"""

    HOST: str = "localhost"
    PORT: int = Field(19530, gt=0, le=65535)
    COLLECTION: str
    COLLECTION_HISTORY_LEN: int = 10
    SEARCH_PARAMS: dict = {"metric_type": "IP", "params": {}}
    INDEX_PARAMS: dict = {
        "index_type": "FLAT",
        "field_name": "vector",
        "metric_type": "IP",
        "params": {},
    }
    USER: str
    PASSWORD: str
    SECURED: bool = False

    @validator("SEARCH_PARAMS", "INDEX_PARAMS", pre=True)
    @classmethod
    # pylint: disable-next=R0801
    def parse_json(cls, v):
        """validation for json"""
        if isinstance(v, str):
            return json.loads(v)
        return v



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

"""containing the DVCStep sending embedding data into Qdrant"""

# pylint: disable=duplicate-code
import itertools
from logging import getLogger

from pandera.typing import DataFrame
from qdrant_client import QdrantClient, models

from wurzel.exceptions import CustomQdrantException, StepFailed
from wurzel.step import TypedStep, step_history
from wurzel.steps.embedding.data import EmbeddingResult
from wurzel.utils import HAS_TLSH

from .data import QdrantResult
from .settings import QdrantSettings

log = getLogger(__name__)


def _batch(iterable, size):
    it = iter(iterable)
    while item := list(itertools.islice(it, size)):
        yield item


class QdrantConnectorStep(
    TypedStep[QdrantSettings, DataFrame[EmbeddingResult], DataFrame[QdrantResult]]
):
    """
    Qdrant connector step. It consumes embedding csv files, creates a new schema and inserts the embeddings
    """

    _timeout: int = 20
    s: QdrantSettings
    client: QdrantClient
    collection_name: str

    def __init__(self) -> None:
        super().__init__()
        # Qdrant stuff passed as environment
        # because we need to enject them into the DVC step during runtime,
        # not during DVC pipeline definition time
        # uri = ":memory:"
        log.info(f"connecting to {self.settings.URI}")
        if not self.settings.APIKEY:
            log.warning(
                "QDRANT__APIKEY for Qdrant not provided."
                " Thus running in non-credential Mode"
            )
        self.client = QdrantClient(
            location=self.settings.URI,
            api_key=self.settings.APIKEY,
            timeout=self._timeout,
        )
        self.collection_name = self.__construct_next_collection_name()
        self.id_iter = self.__id_gen()

    def __del__(self):
        if getattr(self, "client", None):
            self.client.close()

    def finalize(self) -> None:
        self._create_indices()
        self._update_alias()
        self._retire_collections()
        return super().finalize()

    def __id_gen(self):
        i = 0
        while True:
            i += 1
            yield i

    def run(self, inpt: DataFrame[EmbeddingResult]) -> DataFrame[QdrantResult]:
        if not self.client.collection_exists(self.collection_name):
            self._create_collection(len(inpt["vector"].loc[0]))
        df_result = self._insert_embeddings(inpt)
        return df_result

    def _create_collection(self, size: int):
        log.debug(f"Creating Qdrant collection {self.collection_name}")
        self.client.create_collection(
            collection_name=self.collection_name,
            vectors_config=models.VectorParams(
                size=size, distance=self.settings.DISTANCE
            ),
            replication_factor=self.settings.REPLICATION_FACTOR,
        )

    def _insert_embeddings(self, data: DataFrame[EmbeddingResult]):
        log.info(
            "Inserting embeddings",
            extra={"count": len(data), "collection": self.collection_name},
        )
        points = [
            models.PointStruct(
                id=next(self.id_iter),  # type: ignore[arg-type] # No MultiIndex, so always int.
                vector=row["vector"],
                payload={
                    "url": row["url"],
                    "text": row["text"],
                    **self.get_available_hashes(row["text"]),
                    "keywords": row["keywords"],
                    "history": str(step_history.get()),
                },
            )
            for _, row in data.iterrows()
        ]
        for point_chunk in _batch(points, self.settings.BATCH_SIZE):
            operation_info = self.client.upsert(
                collection_name=self.collection_name, wait=True, points=point_chunk
            )
            if operation_info.status != "completed":
                raise StepFailed(
                    f"Failed to insert df chunk into collection '{self.collection_name}' {operation_info}"
                )
            log.info(
                "Successfully inserted vector_chunk",
                extra={"collection": self.collection_name, "count": len(point_chunk)},
            )
        data = [
            {
                **entry.payload,
                "vector": entry.vector,
                "collection": self.collection_name,
                "id": entry.id,
            }
            for entry in points
        ]
        result = DataFrame[QdrantResult](data)
        return result

    def _create_indices(self):
        self.client.create_payload_index(
            collection_name=self.collection_name,
            field_name="keywords",
            field_schema=models.TextIndexParams(
                type=models.TextIndexType.TEXT,
                tokenizer=models.TokenizerType.WHITESPACE,
            ),
        )
        self.client.create_payload_index(
            collection_name=self.collection_name,
            field_name="url",
            field_schema=models.TextIndexParams(
                type=models.TextIndexType.TEXT,
                tokenizer=models.TokenizerType.PREFIX,
                min_token_len=3,
            ),
        )
        self.client.create_payload_index(
            collection_name=self.collection_name,
            field_name="text",
            field_schema=models.TextIndexParams(
                type=models.TextIndexType.TEXT,
                tokenizer=models.TokenizerType.MULTILINGUAL,
            ),
        )
        self.client.create_payload_index(
            collection_name=self.collection_name,
            field_name="history",
            field_schema=models.TextIndexParams(
                type=models.TextIndexType.TEXT, tokenizer=models.TokenizerType.WORD
            ),
        )

    def _retire_collections(self):
        collections_versioned: dict[int, str] = self._get_collection_versions()
        to_delete = list(collections_versioned.keys())[
            : -self.settings.COLLECTION_HISTORY_LEN
        ]
        if not to_delete:
            return

        for col_v in to_delete:
            col = collections_versioned[col_v]
            log.info(f"deleting {col} collection caused by retirement")
            self.client.delete_collection(col)

    def _update_alias(self):
        success = self.client.update_collection_aliases(
            change_aliases_operations=[
                models.CreateAliasOperation(
                    create_alias=models.CreateAlias(
                        collection_name=self.collection_name,
                        alias_name=self.settings.COLLECTION,
                    )
                )
            ]
        )
        if not success:
            raise CustomQdrantException("Alias Update failed")

    def __construct_next_collection_name(self) -> str:
        previous_collections = self._get_collection_versions()
        if not previous_collections:
            return f"{self.settings.COLLECTION}_v1"
        previous_version = max(previous_collections.keys())
        log.info(f"Found version v{previous_version}")
        return f"{self.settings.COLLECTION}_v{previous_version + 1}"

    def _get_collection_versions(self) -> dict[int, str]:
        previous_collections = self.client.get_collections().collections
        versioned_collections = {
            int(previous.name.split("_v")[-1]): previous.name
            for previous in previous_collections
            if f"{self.settings.COLLECTION}_v" in previous.name
        }
        return dict(sorted(versioned_collections.items()))

    @staticmethod
    def get_available_hashes(text: str, encoding: str = "utf-8") -> dict:
        """Compute `n` hashes for a given input text based.

        The number `n` depends on the optionally installed python libs.
        For now only TLSH (Trend Micro Locality Sensitive Hash) is supported

        ## TLSH
        Given a byte stream with a minimum length of 50 bytes TLSH generates a hash value which can be used for similarity comparisons.

        Args:
            text (str): Input text
            encoding (str, optional): Input text will encoded to bytes using this encoding. Defaults to "utf-8".

        Returns:
            dict[str, str]: keys: `text_<algo>_hash` hash as string ! Dict might be empty!
        """
        hashes = {}
        if HAS_TLSH:
            # pylint: disable=no-name-in-module, import-outside-toplevel
            from tlsh import hash as tlsh_hash

            hashes["text_tlsh_hash"] = tlsh_hash(text.encode(encoding))
        return hashes



# SPDX-FileCopyrightText: 2025 Deutsche Telekom AG (opensource@telekom.de)
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

from wurzel.datacontract import MarkdownDataContract
from wurzel.step import Settings, TypedStep


class ManualMarkdownSettings(Settings):
    """Settings fro ManMdstep"""

    FOLDER_PATH: Path


class ManualMarkdownStep(
    TypedStep[ManualMarkdownSettings, None, list[MarkdownDataContract]]
):
    """Data Source for md files from a configurable path"""

    def run(self, inpt: None) -> list[MarkdownDataContract]:
        return [
            MarkdownDataContract.from_file(fp, url_prefix=self.__class__.__name__ + "/")
            for fp in self.settings.FOLDER_PATH.rglob("*.md")
        ]
